/*
 *  array.c
 *
 *  Code borrowed from fs/proc/array.c:
 *
 *  Copyright (C) 1992  by Linus Torvalds
 *  based on ideas by Darren Senn
 *
 * Fixes:
 * Michael. K. Johnson: stat,statm extensions.
 *                      <johnsonm@stolaf.edu>
 *
 * Pauline Middelink :  Made cmdline,envline only break at '\0's, to
 *                      make sure SET_PROCTITLE works. Also removed
 *                      bad '!' which forced address recalculation for
 *                      EVERY character on the current page.
 *                      <middelin@polyware.iaf.nl>
 *
 * Danny ter Haar    :  added cpuinfo
 *                      <dth@cistron.nl>
 *
 * Alessandro Rubini :  profile extension.
 *                      <rubini@ipvvis.unipv.it>
 *
 * Jeff Tranter      :  added BogoMips field to cpuinfo
 *                      <Jeff_Tranter@Mitel.COM>
 *
 * Bruno Haible      :  remove 4K limit for the maps file
 *                      <haible@ma2s2.mathematik.uni-karlsruhe.de>
 *
 * Yves Arrouye      :  remove removal of trailing spaces in get_array.
 *                      <Yves.Arrouye@marin.fdn.fr>
 *
 * Jerome Forissier  :  added per-CPU time information to /proc/stat
 *                      and /proc/<pid>/cpu extension
 *                      <forissier@isia.cma.fr>
 *                      - Incorporation and non-SMP safe operation
 *                      of forissier patch in 2.1.78 by
 *                      Hans Marcus <crowbar@concepts.nl>
 *
 * aeb@cwi.nl        :  /proc/partitions
 *
 *
 * Alan Cox          :  security fixes.
 *                      <Alan.Cox@linux.org>
 *
 * Al Viro           :  safe handling of mm_struct
 *
 * Gerhard Wichert   :  added BIGMEM support
 * Siemens AG           <Gerhard.Wichert@pdb.siemens.de>
 *
 * Al Viro & Jeff Garzik :  moved most of the thing into base.c and
 *                       :  proc_misc.c. The rest may eventually go into
 *                       :  base.c too.
 */


#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#define __NO_VERSION__      /* This isn't "the" file 
                             * of the kernel module */
#include <linux/module.h>
#include <linux/version.h>

#include <linux/autoconf.h>
#include <linux/sched.h>

#include <linux/fs.h>
#include "bklpreempt.h"


static const char *task_state_array[] = {
        "R (running)",          /*  0 */
        "S (sleeping)",         /*  1 */
        "D (disk sleep)",       /*  2 */
        "T (stopped)",          /*  4 */
        "T (tracing stop)",     /*  8 */
        "Z (zombie)",           /* 16 */
        "X (dead)"              /* 32 */
};

static inline const char * get_task_state(struct task_struct *tsk)
{
        unsigned int state = (tsk->state & (TASK_RUNNING |
                                            TASK_INTERRUPTIBLE |
                                            TASK_UNINTERRUPTIBLE |
                                            TASK_STOPPED |
                                            TASK_TRACED)) |
                        (tsk->exit_state & (EXIT_ZOMBIE |
                                            EXIT_DEAD));
        const char **p = &task_state_array[0];

        while (state) {
                p++;
                state >>= 1;
        }
        return *p;
}

char msgbuf[4096];
char xtmp[4096];

/* Dump out relevent info on the given waitqueue - similar to
   do_task_stat() or task_state() in fs/proc/array.c */
int
dump_wait_queue(wait_queue_head_t *h, char *header)
{
	struct list_head *tmp;
	unsigned long flags;
	int cnt = 0;
	struct task_struct *task;

	sprintf(msgbuf, "QUEUE %p: ( ", h);

	spin_lock_irqsave(&h->lock, flags);
	list_for_each(tmp, &h->task_list) {
	    wait_queue_t *curr = list_entry(tmp, wait_queue_t, task_list);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
	    task = curr->private;
#else
	    task = curr->task;
#endif
	    if (cnt)
		sprintf(xtmp, ", %p %s ", task, get_task_state(task));
	    else
		sprintf(xtmp, "%p %s ", task, get_task_state(task));
	    cnt++;
	    strcat(msgbuf, xtmp);
	}
	spin_unlock_irqrestore(&h->lock, flags); 
	sprintf(xtmp, ") ### %s", header);
	strcat(msgbuf, xtmp);
	printk("%s\n", msgbuf);

	return(cnt);
}
