/*
 *  bklpreempt.c
 *
 * Copyright 2009. Quantum Corporation. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write the Free Software Foundation,
 * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/kernel.h>
#include <linux/version.h>
#if defined(__x86_64__)
#include <asm/mmzone.h>
#endif
#include <asm/page.h>
#include <linux/smp_lock.h>

#include "bklpreempt.h"

static int bklbug_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long data);
static int bklbug_waker(void *data);

struct file_operations bklbug_fops = {
	.ioctl = bklbug_ioctl,
	.owner = THIS_MODULE
};

int sleepers = 0;
struct task_struct *trickle_task = NULL;
spinlock_t lock;

static DECLARE_WAIT_QUEUE_HEAD (bklbug_waitqueue);

static int bklbug_sleeper(void *data)
{
	DECLARE_WAITQUEUE (wait, current);

	spin_lock(&lock);
	sleepers++;

	/* This registers the thread that was woken */
	wait.func = bklbug_wake_up;

	add_wait_queue_exclusive (&bklbug_waitqueue, &wait);

	spin_unlock(&lock);

	printk("%p: Going to sleep\n", current);

	set_current_state (TASK_INTERRUPTIBLE);
	schedule ();

	printk("%p: Waking up\n", current);

	set_current_state(TASK_RUNNING);
	remove_wait_queue(&bklbug_waitqueue, &wait);

	return 0;
}


static int bklbug_trickle(void *data)
{
	uint64_t i;
	void *woken;
	for(;;) {

		if (kthread_should_stop()) {
			break;
		}

		msleep(1500);

		spin_lock(&lock);
		lock_kernel();
		if(sleepers) {
			printk("Waking sleeper 1\n");
			sleepers--;
			record_waiter();
			wake_up(&bklbug_waitqueue);
			woken = remove_waiter();
			printk("Woke sleeper %p\n", woken);

			/* Busy spin long enough for the just woken sleeper
			   to get hung up on the BKL */
			for (i = 0; i < 1000000000UL; i++) {
				if (i == 100000000UL)
					dump_wait_queue(&bklbug_waitqueue, "bkl trickle");
			}
			

		}

		if(sleepers) {
			printk("Waking sleeper 2\n");
			sleepers--;
			record_waiter();
			wake_up(&bklbug_waitqueue);
			woken = remove_waiter();
			printk("Woke sleeper %p\n", woken);
		}
		unlock_kernel();
		spin_unlock(&lock);

	}
	return 0;
}

/* Called directly via ioctl to wake up sleepers (even if sleepers is 0) */
static int bklbug_waker(void *data)
{
	spin_lock(&lock);

	if(sleepers) {
		sleepers--;
		wake_up(&bklbug_waitqueue);
	} else {
		wake_up(&bklbug_waitqueue);
	}

	spin_unlock(&lock);
	return 0;
}

static int bklbug_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long data)
{
	int error = 0;

	switch(cmd) {
	case DUMP_WAITQUEUE:
		dump_wait_queue(&bklbug_waitqueue, "dump waitqueue ioctl");
		break;
	case START_SLEEPER:
		bklbug_sleeper(NULL);
		break;
	case WAKEUP_SLEEPER:
		bklbug_waker(NULL);
		break;
	}

	return(error);
}

static __init int bklbug_init(void)
{
	int ret;

	ret = register_chrdev(BKLBUG_MAJOR_NUM, "bklbug", &bklbug_fops);
	printk ("Registered bklbug tester module!\n");

	spin_lock_init(&lock);

	trickle_task = kthread_run(bklbug_trickle, NULL, "bkl_trickle");

	return(ret >= 0 ? 0 : -1);
}

void bklbug_uninit(void)
{
	if(trickle_task) {
		kthread_stop(trickle_task);
	}

	unregister_chrdev(BKLBUG_MAJOR_NUM, "bklbug");
	printk("Unregistered bklbug tester module!\n");
}

module_init(bklbug_init);
module_exit(bklbug_uninit);
